#include "kiss.h"

#include <getopt.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <sys/ioctl.h>
#ifdef __linux__
#include <linux/if.h>
#include <linux/if_tun.h>
#include <net/ethernet.h>
#else
#include <net/if_tun.h>
#endif
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <string.h>

#define SOCKPATH_MAX ( \
      sizeof(struct sockaddr_un) - offsetof(struct sockaddr_un, sun_path) - 1 \
    )

#define DEFAULT_ADDR "::1"
#define DEFAULT_PORT "8001"

#define MTU 1500
#define MAX_PKTLEN (MTU + 4)
#define MAX_FRAMELEN 65536

static struct option longopts[] = {
  { "help",     no_argument,       NULL, 'h' },
  { "tun",      required_argument, NULL, 't' },
  { "hostcall", required_argument, NULL, 'c' },
  { "peercall", required_argument, NULL, 'o' },
  { "socket",   required_argument, NULL, 'l' },
  { "address",  required_argument, NULL, 'a' },
  { "port",     required_argument, NULL, 'p' },
  { NULL, 0, NULL, 0 }
};

static void usage(FILE *out, char *arg0) {
  fprintf(
    out,
    "Usage:%s%s%s\n"
    "  -t <tun device> \\\n"
    "  [-c|--hostcall <own call sign>] \\\n"
    "  [-o|--peercall <call sign of communication partner>] \\\n"
    "  [-l|--socket <path of local socket of KISS server>] \\\n"
    "  [-a|--address <hostname of KISS server>] \\\n"
    "  [-p|--port <port of KISS server> \\\n",
    arg0 ? " " : "",
    arg0 ? arg0 : "",
    arg0 ? " \\" : ""
  );
}

int main(int argc, char **argv) {
  char *tun_filename = NULL;
  char *hostcall = "";
  char *peercall = NULL;
  char *addr = DEFAULT_ADDR;
  char *port = DEFAULT_PORT;
  int tun_fd;
  int kiss_fd;
  {
    int ch;
    while ((ch = getopt_long(
      argc, argv, "ht:c:o:l:a:p:", longopts, NULL
    )) != -1) {
      switch (ch) {
        case 'h': usage(stdout, argc ? argv[0] : NULL); return 0;
        case 't': tun_filename = optarg; break;
        case 'c': hostcall     = optarg; break;
        case 'o': peercall     = optarg; break;
        case 'l':
        case 'a':
          addr = optarg;
          if (ch == 'l') {
            if (strlen(optarg) > SOCKPATH_MAX) {
              fprintf(stderr, "Local socket path too long\n");
              return 1;
            }
            port = 0;
          }
          break;
        case 'p':
          if (!port) {
            fprintf(stderr, "Cannot set port for local socket\n");
            usage(stderr, argc ? argv[0] : NULL);
            return 1;
          }
          port = optarg;
          break;
        default: usage(stderr, argc ? argv[0] : NULL); return 1;
      }
    }
  }
  if (optind != argc || !tun_filename) {
    usage(stderr, argc ? argv[0] : NULL); return 1;
  }
#ifdef __linux__
  tun_fd = open("/dev/net/tun",  O_RDWR | O_NONBLOCK | O_CLOEXEC);
  if (tun_fd == -1) {
    perror("Could not open /dev/net/tun");
    return 1;
  }
  {
    struct ifreq ifr;
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TUN;
    strncpy(ifr.ifr_name, tun_filename, IFNAMSIZ);
    if (ioctl(tun_fd, TUNSETIFF, (void *)&ifr) < 0) {
      perror("Could not initialize TUN device");
      return 1;
    }
  }
#else
  tun_fd = open(tun_filename, O_RDWR | O_NONBLOCK | O_CLOEXEC);
  if (tun_fd == -1) {
    perror("Could not open tun device");
    return 1;
  }
  {
    int option = 1;
    if (ioctl(tun_fd, TUNSIFHEAD, &option) == -1) {
      perror("Could not enable multi-af mode on tun device");
      return 1;
    }
  }
#endif
  if (!port) {
    struct sockaddr_un sockaddr = { .sun_family = AF_LOCAL };
    kiss_fd = socket(
      PF_LOCAL, SOCK_STREAM | SOCK_CLOEXEC, 0
    );
    if (kiss_fd == -1) {
      perror("Could not create local socket");
      return 1;
    }
    strcpy(sockaddr.sun_path, addr);
    if (connect(kiss_fd, (struct sockaddr *)&sockaddr, sizeof(sockaddr))) {
      perror("Could not connect to KISS server via local socket");
      return 1;
    }
  } else {
    struct addrinfo *addrinfo;
    {
      int errcode;
      struct addrinfo *res;
      struct addrinfo hints = {
        .ai_family = AF_UNSPEC,
        .ai_socktype = SOCK_STREAM,
        .ai_protocol = IPPROTO_TCP,
        .ai_flags = AI_ADDRCONFIG | AI_PASSIVE
      };
      errcode = getaddrinfo(addr, port, &hints, &res);
      if (errcode == EAI_SYSTEM) {
        perror("Could not resolve KISS server address");
        return 1;
      } else if (errcode) {
        fprintf(stderr,
          "Could not result KISS server address: %s\n", gai_strerror(errcode)
        );
        return 1;
      }
      for (addrinfo=res; addrinfo; addrinfo=addrinfo->ai_next) {
        if (addrinfo->ai_family == AF_INET6) goto main_address_found;
      }
      for (addrinfo=res; addrinfo; addrinfo=addrinfo->ai_next) {
        if (addrinfo->ai_family == AF_INET) goto main_address_found;
      }
      addrinfo = res;
    }
    main_address_found:
    kiss_fd = socket(
      /* NOTE: not correctly using PF_* but AF_* constants from ai_family */
      addrinfo->ai_family,
      addrinfo->ai_socktype | SOCK_CLOEXEC,
      addrinfo->ai_protocol
    );
    if (kiss_fd == -1) {
      perror("Could not create network socket");
      return 1;
    }
    if (connect(kiss_fd, addrinfo->ai_addr, addrinfo->ai_addrlen)) {
      perror("Could not connect to KISS server");
      return 1;
    }
  }
  {
    int flags;
    flags = fcntl(kiss_fd, F_GETFL, 0);
    if (flags == -1) {
      perror("Could not set socket to KISS server as non-blocking");
      return 1;
    }
    flags |= O_NONBLOCK;
    if (fcntl(kiss_fd, F_SETFL, flags) == -1) {
      perror("Could not set socket to KISS server as non-blocking");
      return 1;
    }
  }
  {
    uint8_t ip_in[MAX_PKTLEN];
    uint8_t ip_out[MAX_PKTLEN];
    uint8_t kissdata_out[MAX_FRAMELEN];
    uint8_t kissframe_out[KISS_MAX_DATAFRAMELEN(MAX_FRAMELEN)];
    uint8_t kiss_in[KISS_MAX_DATAFRAMELEN(MAX_FRAMELEN)];
    ssize_t ip_in_len, kissdata_out_len, kissframe_out_len, kissframe_out_rpos;
    ssize_t kiss_in_len = 0;
    bool kiss_in_ovf = false;
    fd_set readpoll, writepoll, readready, writeready;
    int nfds;
    nfds = (tun_fd > kiss_fd ? tun_fd : kiss_fd) + 1;
    FD_ZERO(&readpoll);
    FD_ZERO(&writepoll);
    FD_SET(tun_fd, &readpoll);
    FD_SET(kiss_fd, &readpoll);
    while (1) {
      readready = readpoll;
      writeready = writepoll;
      pselect(nfds, &readready, &writeready, NULL, NULL, NULL);
      if (FD_ISSET(tun_fd, &readready)) {
        ip_in_len = read(tun_fd, ip_in, MAX_PKTLEN);
        if (ip_in_len >= 0) {
#ifdef __linux__
          uint16_t et;
#else
          uint32_t af;
#endif
          if (ip_in_len < 4) {
            fprintf(stderr, "Packet received from tun device too short\n");
            return 1;
          }
#ifdef __linux__
          et = ntohs(*(uint16_t *)(ip_in + 2));
          if (et == ETHERTYPE_IP || et == ETHERTYPE_IPV6) {
            char *proto = et == ETHERTYPE_IPV6 ? "IPv6" : "IPv4";
#else
          af = ntohl(*(uint32_t *)ip_in);
          if (af == AF_INET || af == AF_INET6) {
            char *proto = af == AF_INET6 ? "IPv6" : "IPv4";
#endif
            if (
              strlen(hostcall) + strlen(proto) + 2 + ip_in_len >
              sizeof(kissdata_out)
            ) {
              fprintf(stderr, "Frame buffer overflow caught (outgoing)\n");
              return 1;
            }
            kissdata_out_len = 0;
            strcpy((char *)kissdata_out+kissdata_out_len, hostcall);
            kissdata_out_len += strlen(hostcall) + 1;
            strcpy((char *)kissdata_out+kissdata_out_len, proto);
            kissdata_out_len += strlen(proto) + 1;
            memcpy(kissdata_out+kissdata_out_len, ip_in+4, ip_in_len-4);
            kissdata_out_len += ip_in_len-4;
            kissframe_out_len = kiss_encode_dataframe(
              kissframe_out, kissdata_out, kissdata_out_len
            );
            kissframe_out_rpos = 0;
            fprintf(stderr, "DEBUG: sending KISS frame\n");
            FD_CLR(tun_fd, &readpoll);
            FD_SET(kiss_fd, &writepoll);
          } else {
#ifdef __linux__
            fprintf(stderr, "Packet with unsupported protocol %i received\n", (int)et);
          }
#else
            fprintf(stderr, "Packet with unsupported address family %i received\n", (int)af);
          }
#endif
        } else if (errno != EINTR) {
          perror("Read error from tun device");
          return 1;
        }
      }
      if (FD_ISSET(kiss_fd, &writeready)) {
        ssize_t written = write(
          kiss_fd,
          kissframe_out+kissframe_out_rpos,
          kissframe_out_len-kissframe_out_rpos
        );
        if (written >= 0) {
          kissframe_out_rpos += written;
          if (kissframe_out_rpos == kissframe_out_len) {
            fprintf(stderr, "DEBUG: KISS frame sent\n");
            FD_CLR(kiss_fd, &writepoll);
            FD_SET(tun_fd, &readpoll);
          }
        } else if (errno != EINTR) {
          perror("Could not send data to KISS server");
          return 1;
        }
      }
      if (FD_ISSET(kiss_fd, &readready)) {
        ssize_t bytes = read(
          kiss_fd,
          kiss_in+kiss_in_len,
          MAX_FRAMELEN-kiss_in_len
        );
        if (bytes >= 0) {
          do {
            int i;
            if (!kiss_in_len) {
              for (i=0; i<bytes; i++) {
                if (kiss_in[i] == 0xC0) {
                  memmove(kiss_in, kiss_in+i, bytes-i);
                  kiss_in_len = 1;
                  bytes--;
                  break;
                }
              }
            }
            if (kiss_in_len || kiss_in_ovf) {
              kiss_in_len += bytes;
              for (i=kiss_in_len-bytes; i<kiss_in_len; i++) {
                if (kiss_in[i] == 0xC0) {
                  if (kiss_in_ovf) {
                    fprintf(stderr, "Oversized KISS frame has been dropped\n");
                    kiss_in_ovf = false;
                    kiss_in_len = 0;
                  } else {
                    if (kiss_in[1] == 0x00) {
                      int i, kiss_content_len, data_len;
                      char *peercall_rcvd;
                      char *proto_rcvd = NULL;
                      uint8_t *data_rcvd = NULL;
                      fprintf(stderr, "KISS data frame received\n");
                      kiss_content_len = kiss_decode(kiss_in);
                      // NOTE: kiss_in[0] is 0x00 now
                      peercall_rcvd = ((char *)kiss_in)+1;
                      for (i=1; i<kiss_content_len; i++) {
                        if (!kiss_in[i]) {
                          i++;
                          proto_rcvd = ((char *)kiss_in)+i;
                          break;
                        }
                      }
                      if (peercall && strcmp(peercall, peercall_rcvd)) {
                        fprintf(stderr, "Peer call does not match\n");
                      } else {
                        for (; i<kiss_content_len; i++) {
                          if (!kiss_in[i]) {
                            i++;
                            data_rcvd = kiss_in+i;
                            data_len = kiss_content_len - i;
                            break;
                          }
                        }
                        if (!data_rcvd) {
                          fprintf(stderr, "No protocol identifier found\n");
                        } else {
                          bool ipv4 = !strcmp(proto_rcvd, "IPv4");
                          bool ipv6 = !strcmp(proto_rcvd, "IPv6");
                          if (!ipv4 && !ipv6) {
                            fprintf(stderr, "Unknown protocol in KISS data\n");
                          } else if (data_len == 0) {
                            fprintf(stderr, "Empty KISS data\n");
                          } else if (data_len > MTU) {
                            fprintf(stderr, "KISS data exceeds MTU\n");
                          } else {
                            ssize_t written;
#ifdef __linux__
                            uint16_t proto = htons(ipv6 ? ETHERTYPE_IPV6 : ETHERTYPE_IP);
                            memset(ip_out, 0, 2);
                            memcpy(ip_out+2, &proto, 4);
#else
                            uint32_t proto = htonl(ipv6 ? AF_INET6 : AF_INET);
                            memcpy(ip_out, &proto, 4);
#endif
                            memcpy(ip_out+4, data_rcvd, data_len);
                            fprintf(stderr, "Sending %s packet\n", ipv6 ? "IPv6" : "IPv4");
                            written = write(tun_fd, ip_out, data_len+4);
                            if (written < 0) {
                              perror("Could not write packet to tun device");
                              return 1;
                            } else if (written < data_len+4) {
                              fprintf(stderr, "Short write to tun device\n");
                              return 1;
                            }
                          }
                        }
                      }
                    }
                    bytes = kiss_in_len - (i+1);
                    memmove(kiss_in, kiss_in+i+1, bytes);
                    kiss_in_len = 0;
                    continue;
                  }
                }
              }
              if (kiss_in_len == sizeof(kiss_in)) {
                fprintf(stderr, "Received KISS frame exceeded maximum size\n");
                kiss_in_ovf = true;
                kiss_in_len = 0;
              }
            }
          } while (0);
        } else if (errno != EINTR) {
          perror("Could not read data from KISS server");
          return 1;
        }
      }
    }
  }
}
