#include "pktfec.h"
#include "generator.c"

#include <math.h>

static void pktfec_encode_block(uint8_t *block) {
  int i, j;
  uint8_t mask;
  const uint8_t *generator = pktfec_generator;
  for (i=PKTFEC_BLOCKDATALEN; i<PKTFEC_BLOCKLEN; i++) block[i] = 0;
  for (i=0; i<PKTFEC_BLOCKDATALEN; i++) {
    for (mask=1<<7; mask; mask>>=1) {
      if (block[i] & mask) {
        for (j=PKTFEC_BLOCKDATALEN; j<PKTFEC_BLOCKLEN; j++) {
          block[j] ^= *generator++;
        }
      } else generator += PKTFEC_BLOCKFECLEN;
    }
  }
}

size_t pktfec_encode(uint8_t *output, const uint8_t *input, size_t length) {
  uint8_t buffer[PKTFEC_BLOCKLEN];
  size_t outputlen, blockcount;
  int lenlen, lastblockdatalen, lastblockfeclen;
  // store length information at end of payload area of first block
  if (length < 0x80) {
    buffer[PKTFEC_BLOCKDATALEN-1] = length;
    lenlen = 1;
  } else if (length <= 0x4080) {
    buffer[PKTFEC_BLOCKDATALEN-1] = 0x80 | ((length-0x80) >> 8);
    buffer[PKTFEC_BLOCKDATALEN-2] = length-0x80;
    lenlen = 2;
  } else if (length <= 0x204080) {
    buffer[PKTFEC_BLOCKDATALEN-1] = 0xc0 | ((length-0x4080) >> 16);
    buffer[PKTFEC_BLOCKDATALEN-2] = (length-0x4080) >> 8;
    buffer[PKTFEC_BLOCKDATALEN-3] = length-0x4080;
    lenlen = 3;
  } else {
    // unsupported length
    return 0;
  }
  // calculate total length of payload for LDPC FEC algorithm
  // (and store it temporarily in outputlen)
  outputlen = lenlen + length + PKTFEC_CRCLEN;
  // calculate total number of blocks needed
  blockcount = (outputlen-1) / PKTFEC_BLOCKDATALEN + 1;
  // handle special case of single block
  if (blockcount == 1) outputlen = PKTFEC_BLOCKLEN;
  else {
    // calculate payload and parity size of last block
    lastblockdatalen = (outputlen-1) % PKTFEC_BLOCKDATALEN + 1;
    lastblockfeclen = PKTFEC_SHORTBLOCKFECLEN(lastblockdatalen);
    // calculate total length of output
    outputlen = (
      (blockcount-1) * PKTFEC_BLOCKLEN +
      lastblockdatalen + lastblockfeclen
    );
  }
  // encoding process (if output is not NULL)
  if (output) {
    size_t outputidx = 0;
    size_t blockidx;
    int byteidx, bitidx;
    uint32_t crc = 0xffffffff;
    int shuffleA, shuffleB, shuffleC;
    // iterate blocks
    for (blockidx=0; blockidx<blockcount; blockidx++) {
      // calculate number of bytes for (used) payload and FEC in block
      int blockdatalen, blockfeclen;
      if (blockidx == 0) {
        // case of first (or single) block
        blockdatalen = PKTFEC_BLOCKDATALEN - lenlen;
        if (blockdatalen > length + PKTFEC_CRCLEN) {
          blockdatalen = length + PKTFEC_CRCLEN;
        }
        //blockfeclen = PKTFEC_BLOCKFECLEN;
      } else if (blockidx == blockcount-1) {
        // case of last (non-single) block
        blockdatalen = lastblockdatalen;
        blockfeclen = lastblockfeclen;
      } else {
        // case of intermediate blocks
        blockdatalen = PKTFEC_BLOCKDATALEN;
        blockfeclen = PKTFEC_BLOCKFECLEN;
      }
      // process blockdatalen bytes
      for (byteidx=0; byteidx<blockdatalen; byteidx++) {
        if (length) {
          // input is still available
          buffer[byteidx] = *input;
          // calculate CRC with most significant bit first
          crc ^= (uint32_t)*input << 24;
          for (bitidx=0; bitidx<8; bitidx++) {
            crc = (crc<<1) ^ ((-((crc&(1<<31))>>31)) & PKTFEC_CRC32POLY);
          }
          // move to next input byte
          input++, length--;
        } else {
          // no more input but CRC pending
          buffer[byteidx] = ~(crc >> 24);
          crc <<= 8;
        }
      }
      // handle first (or single) block differently
      if (blockidx == 0) {
        // fill remaining payload area (excluding length info) with zeros
        for (
          byteidx=blockdatalen;
          byteidx < PKTFEC_BLOCKDATALEN-lenlen;
          byteidx++
        ) {
          buffer[byteidx] = 0;
        }
        // encode block
        pktfec_encode_block(buffer);
        // write block
        for (byteidx=0; byteidx<PKTFEC_BLOCKLEN; byteidx++) {
          output[outputidx++] = buffer[byteidx];
        }
      } else {
        // fill remaining payload area with zeros
        for (byteidx=blockdatalen; byteidx<PKTFEC_BLOCKDATALEN; byteidx++) {
          buffer[byteidx] = 0;
        }
        // encode block
        pktfec_encode_block(buffer);
        // write payload and parity while shuffling
        outputidx = PKTFEC_BLOCKLEN + blockidx - 1;
        shuffleA = 0; shuffleB = -1;
        for (byteidx=0; byteidx<blockdatalen+blockfeclen; byteidx++) {
          if (blockidx == blockcount-1) {
            while (1) {
              shuffleC = (
                shuffleA++ *
                (lastblockdatalen+lastblockfeclen) /
                PKTFEC_BLOCKLEN
              );
              if (shuffleB != shuffleC) {
                shuffleB = shuffleC;
                break;
              }
              outputidx += blockcount-2;
            }
          }
          output[outputidx] = buffer[
            byteidx < blockdatalen ?
            byteidx :
            PKTFEC_BLOCKDATALEN + byteidx - blockdatalen
          ];
          if (blockidx == blockcount-1) {
            outputidx += blockcount-1;
          } else {
            outputidx += blockcount-2;
            shuffleC = (
              shuffleA++ *
              (lastblockdatalen+lastblockfeclen) /
              PKTFEC_BLOCKLEN
            );
            if (shuffleB != shuffleC) {
              shuffleB = shuffleC;
              outputidx++;
            }
          }
        }
      }
    }
  }
  // return length of output
  return outputlen;
}

