#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>


#define QAM_SUCCESS 2

#define QAM_STAGE_SUSPENDED -2
#define QAM_STAGE_FRAMERECOVER -1
#define QAM_STAGE_SYMSYNC 0
#define QAM_STAGE_QAMSYNC 1
#define QAM_STAGE_SYNCED 2


/** Configuration (for modulation and demodulation) **/

typedef struct {
  // configuration values
  int sample_rate;
  int carrier_freq;
  int switch_half_phases;
  int hold_half_phases;
  int bits_per_symbol;
  int qamsync_symbol_count;
  int qamsync_resync_interval;
  int symsync_resync_threshold;
  float tx_volume;
  float tx_delay;
  float symsync_tx_rx_ratio;
  float symsync_tolerance;
  float qamsync_tolerance;
  // derived values
  float carrier_phase_step;
  float switch_phase_step;
  float tx_amplitude;
  float qamsync_phase_precision;
  float symsync_phase_precision;
  int switch_samples;
  int hold_samples;
  int symbol_samples;
  int init_samples;
  int max_amplitude_factor;
  int symsync_rx_count;
  int symsync_tx_half_count;
  int switch_odd;
  int hold_odd;
  bool valid;
} qam_config_t;

#define QAM_DEFAULT_CONFIG { \
    .sample_rate = 48000, \
    .carrier_freq = 1920, \
    .switch_half_phases = 2, \
    .hold_half_phases = 2, \
    .bits_per_symbol = 4, \
    .qamsync_symbol_count = 5, \
    .qamsync_resync_interval = 16, \
    .symsync_resync_threshold = 16, \
    .tx_volume = 0.5, \
    .tx_delay = 0.250, \
    .symsync_tx_rx_ratio = 4, \
    .symsync_tolerance = 4, \
    .qamsync_tolerance = 2, \
  }

bool qam_prepare_config(qam_config_t *config);
/*
 *  Validates and prepares configuration for use. Must be called before using
 *  a configuration for modulation or demodulation. Returns true on success and
 *  false in case of an invalid configuration.
 */


typedef size_t (*qam_writer_fptr)(void *ctx, uint8_t *data, size_t length);
/*
 *  Function pointer to byte writer for modulation. Function may write up to
 *  'length' bytes of audio data pointed to by 'data' and must return the
 *  number of bytes written (zero if no byte could be written). 'ctx' is an
 *  opaque pointer that may be used to support more than one stream per process
 *  and/or remember write errors.
 */

typedef struct {
  qam_config_t *config;
  qam_writer_fptr writer;
  void *writer_ctx;
  int I;
  int Q;
  int phase_odd;
  int sync;
  unsigned int symbol;
  unsigned int bitbuffer;
  int bitcount;
  uint8_t *output;
  size_t output_wpos;
  size_t output_rpos;
  uint8_t *input;
  size_t input_wpos;
  size_t input_rpos;
  int input_finish;
  int stage;
} qam_modulator_t;
/*
 *  Modulator state. Requires additional buffer allocations, thus usually
 *  allocated dynamically using qam_modulator_create().
 */

qam_modulator_t *qam_modulator_create(
  qam_config_t *config,
  qam_writer_fptr writer,
  void *writer_ctx
);
/*
 *  Allocates and initializes a new modulator state.
 *
 *  'config' must point to a configuration struct previously prepared with
 *  qam_prepare_config(). 'writer' must point to a function responsible for
 *  writing the modulated raw audio stream, which gets an opaque pointer
 *  'writer_ctx' passed as first argument on writing.
 *
 *  Returns a pointer to a modulator state which should be free'ed with
 *  qam_modulator_destroy().
 *
 *  On memory allocation error or invalid configuration, returns NULL.
 */

void qam_modulator_reset(qam_modulator_t *state);
/*
 *  Reinitializes modulator state to begin a new packet. Does *not* need to be
 *  called directly after qam_modulator_create() but only when more than one
 *  modulation is to be performed.
 */

void qam_modulator_destroy(qam_modulator_t *state);
/*
 *  Releases all memory associated with the modulator state.
 */

void qam_modulator_send(qam_modulator_t *state, uint8_t *data, size_t length);
/*
 *  Marks a chunk of 'length' bytes (pointed to by 'data') to be modulated.
 *  After each call of qam_modulator_send(), function qam_modulator_run() must
 *  be called repeatedly until modulation is done before qam_modulator_send()
 *  may be called again.
 */

void qam_modulator_finish(qam_modulator_t *state);
/*
 *  Indicates that qam_modulator_send() is not called again, i.e. that the
 *  input data is complete. This will flush buffers on next calls of
 *  qam_modulator_run() and finish the modulation.
 */

bool qam_modulator_run(qam_modulator_t *state);
/*
 *  Performs the actual modulation.
 *
 *  If qam_modulator_finish() has not been called, a return value of true
 *  indicates that more input data can be provided by calling
 *  qam_modulator_send(). Otherwise (i.e. when qam_modulator_finish() had been
 *  called), a return value of true indicates that all data has been modulated.
 *
 *  Returns false if an I/O operation failed (e.g. due to blocking). Then,
 *  qam_modulator_run() needs to be called again when more bytes can be
 *  written. It is responsibility of the caller to not call qam_modulator_run()
 *  in an endless loop if the 'writer' function passed to qam_modulator_init()
 *  encounters an I/O error and always returns zero.
 */

size_t qam_modulate(
  qam_config_t *config,
  float **output, uint8_t *data, size_t length
);
/*
 *  Convenience function which modulates a packet without using a writer
 *  function for the audio but instead returns the complete audio data in a
 *  dynamically allocated array.
 *
 *  'config' must point to a configuration struct previously prepared with
 *  qam_prepare_config(). The whole data to be modulated is pointed to by
 *  'data' and its length in bytes is given by 'length'.
 *
 *  The newly allocated array is stored in the pointer pointed to by 'output',
 *  and its length in blocks of sizeof(float) is given as a return value. In
 *  case of error, zero is returned and the output pointer is set to NULL.
 */

size_t qam_testtone(qam_config_t *config, float **output, int duration);
/*
 *  Creates a test signal consisting of the symbol synchronization tone for
 *  'duration' seconds. If 'duration' is set to 0, creates the shortest
 *  possible repeatable signal.
 *
 *  The newly allocated array is stored in the pointer pointed to by 'output',
 *  and its length in blocks of sizeof(float) is given as a return value. In
 *  case of error, zero is returned and the output pointer is set to NULL.
 */


/** Demodulator **/

typedef size_t (*qam_reader_fptr)(void *ctx, uint8_t *data, size_t length);
/*
 *  Function pointer to byte reader for demodulation. Function may store up to
 *  'length' bytes of read audio data in the buffer pointed to by 'data' and
 *  must return the number of bytes read and stored in the buffer (zero if no
 *  byte could be read). 'ctx' is an opaque pointer that may be used to support
 *  more than one stream per process and/or remember read errors.
 */

typedef struct {
  qam_config_t *config;
  qam_reader_fptr reader;
  void *reader_ctx;
  int phase_odd;
  int stage;
  int adjust;
  int symsync;
  int qamresync;
  int symresync;
  int measurepos;
  float phaseoffset, syncphase;
  float maxamp, zerophase;
  float *maxamps, *zerophases;
  float carrier_detect;
  float *output;
  float *input;
  int input_bytepos;
} qam_demodulator_t;
/*
 *  Demodulator state. Requires additional buffer allocations, thus usually
 *  allocated dynamically using qam_demodulator_create().
 */

qam_demodulator_t *qam_demodulator_create(
  qam_config_t *config,
  qam_reader_fptr reader,
  void *reader_ctx
);
/*
 *  Allocates and initializes a new demodulator state.
 *
 *  'config' must point to a configuration struct previously prepared with
 *  qam_prepare_config(). 'reader' must point to a function responsible for
 *  reading the modulated raw audio stream, which gets an opaque pointer
 *  'reader_ctx' passed as first argument on reading.
 *
 *  Returns a pointer to a demodulator state which should be free'ed with
 *  qam_demodulator_destroy().
 *
 *  On memory allocation error or invalid configuration, returns NULL.
 */

void qam_demodulator_reset(qam_demodulator_t *state);
/*
 *  Resets state of demodulator in such a way that a new packet can be
 *  received.
 *
 *  Does *not* need to be called directly after qam_demodulator_create() but
 *  only to end demodulation of a packet and to wait for a new synchronization.
 *
 *  If the 'reader' function passed to qam_demodulator_create() did not stop
 *  reading at a float-aligned boundary, the demodulator will discard bytes
 *  from the audio input until the alignment is restored.
 */

void qam_demodulator_destroy(qam_demodulator_t *state);
/*
 *  Releases all memory associated with the demodulator state.
 */

bool qam_demodulator_run(qam_demodulator_t *state, int symsync_symbol_count);
/*
 *  Performs a step of the actual demodulation.
 *
 *  The function returns false if no symbol could be demodulated yet. If
 *  'symsync_symbol_count' is not zero, a maximum number of synchronization
 *  symbols will be read until the function returns false. Otherwise only an
 *  I/O error or blocking I/O will cause the function to return false.
 *
 *  Returns true if a symbol could be demodulated. In this case, state->ouput
 *  points to an array of state->config->bits_per_symbol floats between -2.0f
 *  and 2.0f, each corresponding to a bit where a positive float value
 *  indicates a boolean value of 0 and a negative float value indicates a
 *  boolean value of 1. The closer the value to 0.0f, the more uncertainty is
 *  indicated.
 */

void qam_demodulator_suspend(qam_demodulator_t *state);
/*
 *  Causes subsequent calls to qam_demodulator_run() to discard all input bytes
 *  until the demodulator is reset. This can be used to temporarily disable the
 *  demodulator while reading audio. Re-alignment of float boundaries in the
 *  audio stream is ensured by qam_demodulator_reset().
 */

bool qam_demodulator_is_suspended(qam_demodulator_t *state);
/*
 *  Returns true if demodulator has been suspended with
 *  qam_demodulator_suspend().
 */

bool qam_demodulator_is_syncing(qam_demodulator_t *state);
/*
 *  Returns true if demodulator is receiving a valid synchronization signal and
 *  is ready to receive data.
 */

