#ifdef __linux__
#define _GNU_SOURCE
#endif

#include "qam.h"
#include "pktfec.h"
#include "kiss.h"

#include <getopt.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/select.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdbool.h>
#include <math.h>
#include <time.h>
#include <sys/ioctl.h>

#define AUDIOSEND_CMD "play -q -c 1 -b 32 -e float -r 48k -t raw - 2>/dev/null"
#define AUDIORECV_CMD "rec -q -r 48k -e float -b 32 -c 2 -t wav - remix 1 2>/dev/null | sox -q -t wav - -c 1 -b 32 -e float -r 48k -t raw -"
#define RX_VOL_SAMPLES (48000/4)
//#define RX_VOL_SCALE 0.3749975264072418
#define RX_VOL_SCALE 0.5

#define PEER_SENDBUF_MAX 65536
#define PEER_RECVBUF_MAX 65536
#define DEFAULT_PORT "8001"
#define SOCKPATH_MAX ( \
      sizeof(struct sockaddr_un) - offsetof(struct sockaddr_un, sun_path) - 1 \
    )
#define BACKLOG 16
#define DECODER_BUFSIZE (2*8*65536)
#define DECODER_OUTPUT_BUFSIZE 65536

/*
 *  Returns random double precision float between 'minimum' and infinity but
 *  with an average of 'average'.
 */
static double random_duration(double minimum, double average) {
  return (
    minimum - (average - minimum) *
    log(1.0 - (random() & 0x7FFFFFFF) / (double)0x80000000)
  );
}

/*
 *  Struct describing a listener (TCP/IP or local socket).
 */
typedef struct {
  char *address;
  char *port;  // NULL if local socket
  int fd;
} listener_t;

/*
 *  Struct describing a connected client.
 */
typedef struct peer {
  int fd;
  uint8_t sendbuf[PEER_SENDBUF_MAX];
  size_t sendbuf_in, sendbuf_out;
  uint8_t recvbuf[PEER_RECVBUF_MAX];
  size_t recvbuf_in;
  bool recvbuf_ovf;
  /* Linked list: */
  struct peer *prev;
  struct peer *next;
} peer_t;

/*
 *  Helper function to close connection to client and to remove client data
 *  from linked list and free associated memory. Note: This function does not
 *  automatically remove the file descriptor from a set of file descriptors to
 *  poll.
 */
static void close_peer(peer_t **first_peer, peer_t *peer) {
  close(peer->fd);
  if (peer->prev) peer->prev->next = peer->next;
  else *first_peer = peer->next;
  if (peer->next) peer->next->prev = peer->prev;
  free(peer);
}

/*
 *  Context for audio_read() function.
 */
typedef struct {
  int fd;
  int error;
  bool eof;
} audio_read_ctx_t;

/*
 *  Audio reader to be used for demodulator.
 */
static size_t audio_read(void *ctx, uint8_t *input, size_t length) {
  audio_read_ctx_t *const audio_read_ctx = ctx;
  if (!audio_read_ctx->error && !audio_read_ctx->eof) {
    ssize_t bytes = read(audio_read_ctx->fd, input, length);
    if (bytes > 0) return bytes;
    if (!bytes) audio_read_ctx->eof = true;
    else if (errno != EAGAIN && errno != EINTR) audio_read_ctx->error = errno;
  }
  return 0;
}

/*
 *  Struct to store a packet in the sending queue.
 */
typedef struct {
  uint8_t *data;
  size_t length;
} packet_t;

/*
 *  Command line arguments.
 */
static struct option longopts[] = {
  { "help",          no_argument,       NULL, 'h' },
  { "socket",        required_argument, NULL, 'l' },
  { "address",       required_argument, NULL, 'a' },
  { "port",          required_argument, NULL, 'p' },
  { "audiosend",     required_argument, NULL, 's' },
  { "audiorecv",     required_argument, NULL, 'r' },
  { "bits",          required_argument, NULL, 'b' },
  { "freq",          required_argument, NULL, 'f' },
  { "mindelay",      required_argument, NULL, 'm' },
  { "avrdelay",      required_argument, NULL, 'd' },
  { "queue",         required_argument, NULL, 'q' },
  { "duplex",        no_argument,       NULL, 'x' },
  { "peer-to-peer",  no_argument,       NULL, 't' },
  { "volume",        required_argument, NULL, 'V' },
  { "test-vol-rx",   no_argument,       NULL, 'R' },
  { "test-vol-tx",   no_argument,       NULL, 'T' },
  { "ptt-rts-on",    required_argument, NULL, 1 },
  { "ptt-rtr-on",    required_argument, NULL, 1 },
  { "ptt-dtr-on",    required_argument, NULL, 2 },
  { "ptt-rts-off",   required_argument, NULL, 3 },
  { "ptt-rtr-off",   required_argument, NULL, 3 },
  { "ptt-dtr-off",   required_argument, NULL, 4 },
  { "rts-always-on", no_argument,       NULL, 5 },
  { "rtr-always-on", no_argument,       NULL, 5 },
  { "dtr-always-on", no_argument,       NULL, 6 },
  { NULL, 0, NULL, 0 }
};

/*
 *  Prints usage information to a stream.
 */
static void usage(FILE *out, char *arg0) {
  fprintf(
    out,
    "Usage:%s%s%s\n"
    "  [-l|--socket <path of local socket to listen on>] \\\n"
    "  [-a|--address <address of network interface to listen on>] \\\n"
    "  [-p|--port <port to listen on> \\\n"
    "  [-s|--audiosend <shell command accepting and sending single precision floats>] \\\n"
    "  [-r|--audiorecv <shell command receiving and returning single precision floats>] \\\n"
    "  [-b|--bits <bits per symbol, must be even, defaults to 4>] \\\n"
    "  [-f|--freq <carrier frequency, defaults to 1920>] \\\n"
    "  [-m|--mindelay <minimum tx delay after receiving signal, defaults to 1>] \\\n"
    "  [-d|--avrdelay <average tx delay after receiving signal, defaults to 10>] \\\n"
    "  [-q|--queue <length of tx queue, defaults to 3>] \\\n"
    "  [-x|--duplex]\n"
    "  [-t|--peer-to-peer]\n"
    "  [-V|--volume]\n"
    "  [-R|--test-vol-rx]\n"
    "  [-T|--test-vol-tx]\n"
    "  [--ptt-rts-on |--ptt-rtr-on |--ptt-dtr-on|\n"
    "   --ptt-rts-off|--ptt-rtr-off|--ptt-dtr-off <serial device>]\n"
    "  [--rts-always-on|--rtr-always-on|--dtr-always-on]\n",
    arg0 ? " " : "",
    arg0 ? arg0 : "",
    arg0 ? " \\" : ""
  );
}

/*
 *  Helper functions to print errors. Returning 1.
 */
static int static_error(const char *msg) {
  fprintf(stderr, "Error: %s\n", msg);
  return 1;
}
static int static_error2(const char *msg, const char *detail) {
  fprintf(stderr, "Error: %s: %s\n", msg, detail);
  return 1;
}
static int errno_error(const char *msg) {
  fputs("Error: ", stderr);
  perror(msg);
  return 1;
}

/*
 *  Function initializing random generator.
 */
static void random_init() {
  struct timespec now;
  if (clock_gettime(CLOCK_MONOTONIC, &now)) {
    errno_error("Could not access CLOCK_MONOTONIC");
    abort();
  }
  srandom(
    (unsigned long)now.tv_sec * (unsigned long)1e9 +
    (unsigned long)now.tv_nsec
  );
}

/*
 *  Returns timestamp (as double) representing 'duration' seconds in the
 *  future.
 */
static double future(double duration) {
  struct timespec now;
  if (clock_gettime(CLOCK_MONOTONIC, &now)) {
    errno_error("Could not access CLOCK_MONOTONIC");
    abort();
  }
  return now.tv_sec + now.tv_nsec / 1.0e9 + duration;
}

/*
 *  Fills a timespec struct 'dst' with the amount of time a 'timestamp' is in
 *  the future (from now). Fills 'dst' with 0 seconds and 0 nanoseconds if
 *  'timestamp' is in the past.
 */
static void tdiffnow(struct timespec *dst, double timestamp) {
  struct timespec now;
  if (clock_gettime(CLOCK_MONOTONIC, &now)) {
    errno_error("Could not access CLOCK_MONOTONIC");
    abort();
  }
  timestamp -= now.tv_sec + now.tv_nsec / 1.0e9;
  if (timestamp > 0) {
    dst->tv_sec = timestamp;
    dst->tv_nsec = 1e9 * (timestamp - dst->tv_sec);
  } else {
    dst->tv_sec = 0;
    dst->tv_nsec = 0;
  }
}

/*
 *  Returns true if 'timestamp' is no longer in the future.
 */
static bool elapsed(double timestamp) {
  struct timespec now;
  if (clock_gettime(CLOCK_MONOTONIC, &now)) {
    errno_error("Could not access CLOCK_MONOTONIC");
    abort();
  }
  return timestamp <= now.tv_sec + now.tv_nsec / 1.0e9;
}

int main(int argc, char **argv) {

  /* Variables regarding command line arguments */
  double mindelay = 1;
  double avrdelay = 10;
  int queuelen = 3;
  bool duplex = false;
  bool peertopeer = false;
  int ptt_fd = -1;
  int ptt_flag = 0;
  unsigned long ptt_cmd_on;
  unsigned long ptt_cmd_off;
  bool rtr_always = false, dtr_always = false;
  bool test_vol_rx = false;
  bool test_vol_tx = false;
  /* NOTE: listener_cnt = 0 denotes that no listener was specified on
   * command line (yet) */
  int listener_cnt = 0;
  listener_t *listeners;
  char *audiosend_cmd = AUDIOSEND_CMD;
  char *audiorecv_cmd = AUDIORECV_CMD;
  qam_config_t qam_config[1] = { QAM_DEFAULT_CONFIG };
  qam_config->bits_per_symbol = 4;
  qam_config->carrier_freq = 1920;
  listeners = malloc(sizeof(*listeners));
  listeners[0].address = "::1";
  listeners[0].port = DEFAULT_PORT;

  /* Read command line arguments */
  {
    int ch;
    char *endptr;
    long intval;
    while ((ch = getopt_long(
      argc, argv, "hl:a:p:s:r:b:f:m:d:q:xtV:RT", longopts, NULL
    )) != -1) {
      switch (ch) {
      case 'h':
        usage(stdout, argc ? argv[0] : NULL);
        return 0;
      case 'l':
      case 'a':
        if (listener_cnt++) {
          if (listener_cnt > SIZE_MAX / sizeof(*listeners)) {
            return static_error("Out of memory");
          }
          listeners = realloc(listeners, listener_cnt * sizeof(*listeners));
          if (!listeners) {
            return static_error("Out of memory");
          }
        }
        listeners[listener_cnt-1].address = optarg;
        if (ch == 'l') {
          if (strlen(optarg) > SOCKPATH_MAX) {
            return static_error("Local socket path too long");
          }
          listeners[listener_cnt-1].port = 0;
        } else if (listener_cnt > 1) {
          listeners[listener_cnt-1].port = DEFAULT_PORT;
        }
        break;
      case 'p':
        if (!listeners[listener_cnt ? listener_cnt-1 : 0].port) {
          static_error("Cannot set port for local socket");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        listeners[listener_cnt ? listener_cnt-1 : 0].port = optarg;
        break;
      case 's': audiosend_cmd = optarg; break;
      case 'r': audiorecv_cmd = optarg; break;
      case 'b':
        if      (!strcmp(optarg,  "2")) qam_config->bits_per_symbol =  2;
        else if (!strcmp(optarg,  "4")) qam_config->bits_per_symbol =  4;
        else if (!strcmp(optarg,  "6")) qam_config->bits_per_symbol =  6;
        else if (!strcmp(optarg,  "8")) qam_config->bits_per_symbol =  8;
        else if (!strcmp(optarg, "10")) qam_config->bits_per_symbol = 10;
        else if (!strcmp(optarg, "12")) qam_config->bits_per_symbol = 12;
        else if (!strcmp(optarg, "14")) qam_config->bits_per_symbol = 14;
        else if (!strcmp(optarg, "16")) qam_config->bits_per_symbol = 16;
        else {
          static_error("Bits per symbol must be even and between 2 and 16");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        break;
      case 'f':
        if      (!strcmp(optarg,  "1500")) qam_config->carrier_freq = 1500;
        else if (!strcmp(optarg,  "1600")) qam_config->carrier_freq = 1600;
        else if (!strcmp(optarg,  "1920")) qam_config->carrier_freq = 1920;
        else if (!strcmp(optarg,  "2000")) qam_config->carrier_freq = 2000;
        else if (!strcmp(optarg,  "2400")) qam_config->carrier_freq = 2400;
        else if (!strcmp(optarg,  "3000")) qam_config->carrier_freq = 3000;
        else if (!strcmp(optarg,  "4000")) qam_config->carrier_freq = 4000;
        else if (!strcmp(optarg,  "4800")) qam_config->carrier_freq = 4800;
        else if (!strcmp(optarg,  "6000")) qam_config->carrier_freq = 6000;
        else {
          static_error("Only supported carrier frequencies are 1500, 1600, 1920, 2000, 2400, 3000, 4000, 4800, 6000");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        break;
      case 'm':
        mindelay = strtod(optarg, &endptr);
        if (*endptr || !(mindelay >= 0 && mindelay <= 3600)) {
          static_error("Illegal minimum delay time specified");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        if (avrdelay < mindelay) avrdelay = mindelay;
        break;
      case 'd':
        avrdelay = strtod(optarg, &endptr);
        if (*endptr || !(avrdelay >= 0 && avrdelay <= 86400)) {
          static_error("Illegal average delay time specified");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        if (avrdelay < mindelay) mindelay = avrdelay;
        break;
      case 'q':
        errno = 0;
        intval = strtol(optarg, &endptr, 10);
        if (*endptr || errno || !(intval >= 1 && intval <= 100000)) {
          static_error("Illegal queue length specified");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        queuelen = intval;
        break;
      case 'x': duplex = true; break;
      case 't': peertopeer = true; break;
      case 'V':
        qam_config->tx_volume = strtod(optarg, &endptr);
        if (*endptr || !(qam_config->tx_volume >= 0)) {
          static_error("Illegal volume specified");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        break;
      case 'R': test_vol_rx = true; break;
      case 'T': test_vol_tx = true; break;
      case 1: case 2: case 3: case 4:
        if (ptt_flag) {
          static_error("Only one PTT option allowed");
          usage(stderr, argc ? argv[0] : NULL);
          return 1;
        }
        ptt_fd = open(optarg, O_RDWR | O_NONBLOCK | O_CLOEXEC);
        if (ptt_fd < 0) return errno_error("Could not open serial device");
        if (ch == 1 || ch == 2) {
          ptt_cmd_on = TIOCMBIS; ptt_cmd_off = TIOCMBIC;
        } else if (ch == 3 || ch == 4) {
          ptt_cmd_on = TIOCMBIC; ptt_cmd_off = TIOCMBIS;
        } else {
          abort();
        }
        if      (ch == 1 || ch == 3) ptt_flag = TIOCM_RTS;
        else if (ch == 2 || ch == 4) ptt_flag = TIOCM_DTR;
        else abort();
        break;
      case 5: rtr_always = true; break;
      case 6: dtr_always = true; break;
      default:
        usage(stderr, argc ? argv[0] : NULL);
        return 1;
      }
    }
    if (optind != argc) {
      usage(stderr, argc ? argv[0] : NULL);
      return 1;
    }
    if (!listener_cnt) listener_cnt = 1;
  }
  if (rtr_always && ptt_flag == TIOCM_RTS) {
    return static_error("Conflicting options for RTR/RTS line");
  }
  if (dtr_always && ptt_flag == TIOCM_DTR) {
    return static_error("Conflicting options for DTR line");
  }

  /* Set handshake lines */
  if (ptt_flag) {
    int i;
    for (i=0; i<=20; i++) {
      if (i) usleep(10000);  // NOTE: delay needed after opening serial device
      if (ioctl(ptt_fd, ptt_cmd_off, &ptt_flag) < 0) {
        return errno_error("Could not disable PTT");
      }
      if (!(ptt_flag & TIOCM_RTS)) {
        int flag = TIOCM_RTS;
        if (rtr_always) {
          if (ioctl(ptt_fd, TIOCMBIS, &flag) < 0) {
            return errno_error("Could not switch RTR/RTS on");
          }
        } else {
          if (ioctl(ptt_fd, TIOCMBIC, &flag) < 0) {
            return errno_error("Could not switch RTR/RTS off");
          }
        }
      }
      if (!(ptt_flag & TIOCM_DTR)) {
        int flag = TIOCM_DTR;
        if (dtr_always) {
          if (ioctl(ptt_fd, TIOCMBIS, &flag) < 0) {
            return errno_error("Could not switch DTR on");
          }
        } else {
          if (ioctl(ptt_fd, TIOCMBIC, &flag) < 0) {
            return errno_error("Could not switch DTR off");
          }
        }
      }
    }
  }

  if (test_vol_rx) {
    float samples[RX_VOL_SAMPLES];
    float energy;
    float vol;
    int i;
    FILE *f = popen(audiorecv_cmd, "r");
    if (!f) {
      return static_error("Could not open audio receive command");
    }
    while (1) {
      size_t bytes = fread(samples, sizeof(float), RX_VOL_SAMPLES, f);
      if (bytes < RX_VOL_SAMPLES) {
        return static_error("Could not read further audio data");
      }
      energy = 0;
      for (i=0; i<RX_VOL_SAMPLES; i++) energy += samples[i] * samples[i];
      vol = log10(energy / ((double)RX_VOL_SCALE * RX_VOL_SAMPLES)) * 10;
      if (vol < -99.9) vol = -1.0/0.0;
      fprintf(stdout, "RX volume: %5.1f dB\n", vol);
    }
  }

  /* Prepare QAM configuration */
  if (!qam_prepare_config(qam_config)) abort();

  /* Initiatialize random generator */
  random_init();

  /* Main program */
  {
    int i;

    /* File descriptors to be polled for reading/writing and
     * file descriptors being ready for reading/writing. */
    fd_set readpoll, writepoll, readready, writeready;

    /* Linked list of peers and pointer to one peer */
    peer_t *first_peer = NULL;
    peer_t *peer;

    /* Variables related to demodulator and decoder */
    audio_read_ctx_t audio_read_ctx = { 0, };
    qam_demodulator_t *demodulator;
    pktfec_decoder_t decoder;
    float decoder_buf[DECODER_BUFSIZE];
    uint8_t decoder_outbuf[DECODER_OUTPUT_BUFSIZE];
    int decoder_bufpos = 0;
    int decoder_needed = 8*PKTFEC_BLOCKLEN;
    int decoder_state = 0;

    /* Macro for demodulator and decoder reset */
#define main_demodulator_reset() do { \
      qam_demodulator_reset(demodulator); \
      decoder_bufpos = 0; \
      decoder_needed = 8*PKTFEC_BLOCKLEN; \
      decoder_state = 0; \
    } while (0)

    /* Variables related to send queue */
    packet_t *queue;
    int queuecnt = 0;

    /* Variables related to sending process */
    FILE *audiosend_proc = NULL;
    int audiosend_fd;
    uint8_t *audiosend_next_data = NULL;
    size_t audiosend_next_datalen;
    uint8_t *audiosend_data = NULL;
    size_t audiosend_datalen;
    size_t audiosend_done;

    /* Helper macro for sending process */
#define main_audiosend_next() do { \
      audiosend_data = audiosend_next_data; \
      audiosend_datalen = audiosend_next_datalen; \
      audiosend_done = 0; \
      audiosend_next_data = NULL; \
    } while (0)

    /* Variables related to TX blocking on incoming synchronization */
    double blocked_until = 0;
    bool blocked = false;
    bool temp_blocked = false;
    bool any_demodulated;
    struct timespec delay;

    /* Allocate array for queue */
    queue = malloc(queuelen * sizeof(packet_t));
    if (!queue) return static_error("Out of memory");

    /* Initialize file descriptor sets used for polling */
    FD_ZERO(&readpoll);
    FD_ZERO(&writepoll);

    /* Initialize listeners */
    for (i=0; i<listener_cnt; i++) {
      listener_t *listener = listeners + i;
      if (!listener->port) {
        struct stat sb;
        struct sockaddr_un sockaddr = { .sun_family = AF_LOCAL };
        if (stat(listener->address, &sb) == 0) {
          if (S_ISSOCK(sb.st_mode)) unlink(listener->address);
        }
        listener->fd = socket(
          PF_LOCAL, SOCK_STREAM | SOCK_CLOEXEC | SOCK_NONBLOCK, 0
        );
        if (listener->fd < 0) {
          return errno_error("Could not create local socket");
        }
        strcpy(sockaddr.sun_path, listener->address);
        if (bind(
          listener->fd, (struct sockaddr *)&sockaddr, sizeof(sockaddr)
        )) {
          return errno_error("Could not bind local socket");
        }
        if (listen(listener->fd, BACKLOG)) {
          return errno_error("Could not listen on local socket");
        }
      } else {
        struct addrinfo *addrinfo;
        {
          int errcode;
          struct addrinfo *res;
          struct addrinfo hints = {
            .ai_family = AF_UNSPEC,
            .ai_socktype = SOCK_STREAM,
            .ai_protocol = IPPROTO_TCP,
            .ai_flags = AI_ADDRCONFIG | AI_PASSIVE
          };
          errcode = getaddrinfo(listener->address, listener->port, &hints, &res);
          if (errcode == EAI_SYSTEM) {
            return errno_error("Could not resolve local network address");
          } else if (errcode) return static_error2(
            "Could not resolve local network address", gai_strerror(errcode)
          );
          for (addrinfo=res; addrinfo; addrinfo=addrinfo->ai_next) {
            if (addrinfo->ai_family == AF_INET6) goto main_address_found;
          }
          for (addrinfo=res; addrinfo; addrinfo=addrinfo->ai_next) {
            if (addrinfo->ai_family == AF_INET) goto main_address_found;
          }
          addrinfo = res;
        }
        main_address_found:
        listener->fd = socket(
          /* NOTE: not correctly using PF_* but AF_* constants from ai_family */
          addrinfo->ai_family,
          addrinfo->ai_socktype | SOCK_CLOEXEC | SOCK_NONBLOCK,
          addrinfo->ai_protocol
        );
        if (listener->fd < 0) {
          return errno_error("Could not create network socket");
        }
        {
          static const int reuseval = 1;
          if (setsockopt(
            listener->fd, SOL_SOCKET, SO_REUSEADDR, &reuseval, sizeof(reuseval)
          )) {
            return errno_error("Could not set SO_REUSEADDR with setsockopt call");
          }
        }
        if (bind(listener->fd, addrinfo->ai_addr, addrinfo->ai_addrlen)) {
          return errno_error("Could not bind network socket to local address");
        }
        if (listen(listener->fd, BACKLOG)) {
          return errno_error("Could not listen on network socket");
        }
      }
      FD_SET(listener->fd, &readpoll);
    }

    /* Initialize facilities for reading and processing audio */
    {
      FILE *audiorecv_proc = popen(audiorecv_cmd, "r");
      if (!audiorecv_proc) {
        return static_error("Could not open audio receive command");
      }
      audio_read_ctx.fd = fileno(audiorecv_proc);
      if (fcntl(audio_read_ctx.fd, F_SETFL, O_NONBLOCK)) {
        return errno_error("Cannot set audio receive socket non-blocking");
      };
      FD_SET(audio_read_ctx.fd, &readpoll);
    }
    demodulator = qam_demodulator_create(
      qam_config,
      audio_read,
      (void *)&audio_read_ctx
    );
    if (!demodulator) {
      return static_error("Could not initialize demodulator");
    }

    /* Create test tone, if requested */
    if (test_vol_tx) {
      audiosend_next_datalen = sizeof(float) * qam_testtone(
        qam_config, (float **)&audiosend_next_data, 2
      );
      if (!audiosend_next_data) return static_error("Out of memory");
    }

    /* Main loop */
    while (1) {

      /* Wait */
      if (blocked) tdiffnow(&delay, blocked_until);
      readready = readpoll; writeready = writepoll;
      pselect(FD_SETSIZE, &readready, &writeready, NULL, NULL, NULL);
      if (blocked && elapsed(blocked_until)) {
        fprintf(stderr, "TX block lifted\n");
        blocked = false;
      }

      /* Handle new connections */
      for (i=0; i<listener_cnt; i++) {
        listener_t *const listener = listeners + i;
        int peerfd;
        do {
          peerfd = accept4(
            listener->fd, NULL, NULL, SOCK_CLOEXEC | SOCK_NONBLOCK
          );
        } while (peerfd < 0 && errno == EINTR);
        if (peerfd >= 0) {
          peer = calloc(1, sizeof(peer_t));
          if (!peer) return static_error("Out of memory");
          peer->fd = peerfd;
          peer->next = first_peer;
          if (first_peer) first_peer->prev = peer;
          first_peer = peer;
          FD_SET(peer->fd, &readpoll);
          if (listener->port) {
            struct sockaddr_in6 name;
            socklen_t namelen = sizeof(struct sockaddr_in6);
            char addrstrbuf[INET6_ADDRSTRLEN];
            const char *addrstr;
            if (getpeername(peer->fd, (struct sockaddr *)&name, &namelen)) {
              return errno_error("Could not get peer address");
            }
            addrstr = inet_ntop(
              AF_INET6,
              name.sin6_addr.s6_addr,
              addrstrbuf,
              sizeof(addrstrbuf)
            );
            if (!addrstr) {
              return errno_error("Could not format peer address");
            }
            fprintf(stderr, "Connection from %s\n", addrstr);
          } else {
            fprintf(stderr, "Connection on local socket\n");
          }
        }
      }

      /* handle incoming audio data */
      any_demodulated = false;
      while (qam_demodulator_run(demodulator, 0)) {
        int inputpos = 0;
        any_demodulated = true;
        while (1) {
          while (
            decoder_bufpos < decoder_needed &&
            inputpos < qam_config->bits_per_symbol
          ) {
            decoder_buf[decoder_bufpos++] = demodulator->output[inputpos++];
          }
          if (decoder_bufpos < decoder_needed) break;
          if (decoder_state == 0) {
            if (
              pktfec_decode_start(
                &decoder, decoder_buf, 8*PKTFEC_BLOCKLEN
              ) &&
              8*decoder.extralen <= DECODER_BUFSIZE
            ) {
              decoder_bufpos = 0;
              decoder_needed = 8*decoder.extralen;
              decoder_state = 1;
            } else {
              main_demodulator_reset();
              break;
            }
          } else if (decoder_state == 1) {
            if (
              decoder.outputlen <= DECODER_OUTPUT_BUFSIZE &&
              pktfec_decode_finish(
                &decoder,
                decoder_outbuf,
                decoder_buf,
                decoder_needed
              )
            ) {
              fprintf(stderr, "####### Decoded audio signal ######\n");
              if (peertopeer) blocked = false;
              if (
                kiss_encode_dataframe(NULL, decoder_outbuf, decoder.outputlen)
                <= PEER_SENDBUF_MAX
              ) {
                for (peer=first_peer; peer; peer=peer->next) {
                  if (!peer->sendbuf_in) {
                    // NOTE: Only one KISS frame is buffered
                    peer->sendbuf_in = kiss_encode_dataframe(
                      peer->sendbuf, decoder_outbuf, decoder.outputlen
                    );
                    FD_SET(peer->fd, &writepoll);
                  }
                }
              }
            }
            main_demodulator_reset();
            break;
          }
        }
      }
      if (!duplex) {
        if (peertopeer) {
          temp_blocked = qam_demodulator_is_syncing(demodulator);
        } else if (any_demodulated || qam_demodulator_is_syncing(demodulator)) {
          if (blocked == false) {
            fprintf(stderr, "Incoming signal, blocking TX\n");
          }
          blocked_until = future(random_duration(mindelay, avrdelay));
          blocked = true;
        }
      }
      if (audio_read_ctx.error) {
        errno = audio_read_ctx.error;
        return errno_error("Could not read audio stream");
      }
      if (audio_read_ctx.eof) return static_error("Audio stream ended");

      /* Handle peers */
      for (peer=first_peer; peer; peer=peer->next) {
        /* Receiving data from peer */
        {
          ssize_t bytes = read(
            peer->fd,
            peer->recvbuf+peer->recvbuf_in,
            PEER_RECVBUF_MAX-peer->recvbuf_in
          );
          if (bytes < 0) {
            if (errno != EAGAIN && errno != EINTR) {
              fprintf(stderr, "Closing client connection\n");
              FD_CLR(peer->fd, &readpoll);
              FD_CLR(peer->fd, &writepoll);
              close_peer(&first_peer, peer);
              continue;
            }
          } else do {
            int i;
            if (!peer->recvbuf_in) {
              for (i=0; i<bytes; i++) {
                if (peer->recvbuf[i] == 0xC0) {
                  memmove(peer->recvbuf, peer->recvbuf+i, bytes-i);
                  peer->recvbuf_in = 1;
                  bytes--;
                  break;
                }
              }
            }
            if (peer->recvbuf_in || peer->recvbuf_ovf) {
              peer->recvbuf_in += bytes;
              for (i=peer->recvbuf_in-bytes; i<peer->recvbuf_in; i++) {
                if (peer->recvbuf[i] == 0xC0) {
                  if (peer->recvbuf_ovf) {
                    fprintf(stderr, "Frame which exceeded maximum size was dropped\n");
                    peer->recvbuf_ovf = false;
                    peer->recvbuf_in = 0;
                  } else {
                    fprintf(stderr, "Frame received from peer\n");
                    if (peer->recvbuf[1] == 0x00) {
                      if (queuecnt == queuelen) {
                        fprintf(stderr, "Send queue overflow, dropping oldest packet\n");
                        free(queue[0].data);
                        memmove(queue, queue+1, --queuecnt * sizeof(packet_t));
                      }
                      queue[queuecnt].length = kiss_decode(peer->recvbuf) - 1;
                      queue[queuecnt].data = malloc(queue[queuecnt].length);
                      if (!queue[queuecnt].data) {
                        return static_error("Out of memory");
                      }
                      memcpy(
                        queue[queuecnt].data,
                        peer->recvbuf + 1,
                        queue[queuecnt].length
                      );
                      queuecnt++;
                    }
                    /* Handle potential next KISS frames in buffer */
                    bytes = peer->recvbuf_in - (i+1);
                    memmove(peer->recvbuf, peer->recvbuf+i+1, bytes);
                    peer->recvbuf_in = 0;
                    continue;
                  }
                }
              }
              if (peer->recvbuf_in == PEER_RECVBUF_MAX) {
                fprintf(stderr, "Frame received from peer exceeded maximum size\n");
                peer->recvbuf_ovf = true;
                peer->recvbuf_in = 0;
              }
            }
          } while (0);
        }
        /* Sending data to peer */
        if (peer->sendbuf_in) {
          ssize_t written = write(
            peer->fd,
            peer->sendbuf+peer->sendbuf_out,
            peer->sendbuf_in-peer->sendbuf_out
          );
          if (written < 0) {
            if (errno != EAGAIN && errno != EINTR) {
              fprintf(stderr, "Closing client connection\n");
              FD_CLR(peer->fd, &readpoll);
              FD_CLR(peer->fd, &writepoll);
              close_peer(&first_peer, peer);
              continue;
            }
          } else {
            peer->sendbuf_out += written;
            if (peer->sendbuf_out == peer->sendbuf_in) {
              peer->sendbuf_in = 0;
              peer->sendbuf_out = 0;
              FD_CLR(peer->fd, &writepoll);
            }
          }
        }
      }

      /* Process TX queue */
      main_repeat_audiosend:
      if (
        queuecnt && !audiosend_next_data && !blocked && !temp_blocked &&
        (duplex || !peertopeer || !audiosend_data)
      ) {
        uint8_t *encoded;
        size_t encoded_length;
        fprintf(stderr, "Sending audio\n");
        encoded_length = pktfec_encode(NULL, queue[0].data, queue[0].length);
        encoded = malloc(encoded_length);
        if (!encoded) return static_error("Out of memory");
        pktfec_encode(encoded, queue[0].data, queue[0].length);
        free(queue[0].data);
        memmove(queue, queue+1, --queuecnt * sizeof(packet_t));
        audiosend_next_datalen = sizeof(float) * qam_modulate(
          qam_config,
          (float **)&audiosend_next_data, encoded, encoded_length
        );
        if (!audiosend_next_data) return static_error("Out of memory");
        free(encoded);
      }
      if (audiosend_next_data && !audiosend_data) {
        main_audiosend_next();
        audiosend_proc = popen(audiosend_cmd, "w");
        if (!audiosend_proc) {
          return static_error("Could not open audio send command");
        }
        audiosend_fd = fileno(audiosend_proc);
        if (fcntl(audiosend_fd, F_SETFL, O_NONBLOCK)) {
          return errno_error("Could not set audio send socket non-blocking");
        };
        FD_SET(audiosend_fd, &writepoll);
        if (!duplex) qam_demodulator_suspend(demodulator);
        if (ptt_flag && ioctl(ptt_fd, ptt_cmd_on, &ptt_flag) < 0) {
          return errno_error("Could not enable PTT");
        }
      }
      if (test_vol_tx && !audiosend_next_data) {
        audiosend_next_data = audiosend_data;
        audiosend_next_datalen = audiosend_datalen;
      }
      if (audiosend_data) {
        ssize_t written = write(
          audiosend_fd,
          audiosend_data + audiosend_done,
          audiosend_datalen - audiosend_done
        );
        if (written < 0) {
          if (errno != EAGAIN && errno != EINTR) {
            return errno_error("Could not write audio stream");
          }
        } else {
          audiosend_done += written;
          if (audiosend_done == audiosend_datalen) {
            free(audiosend_data);
            if (audiosend_next_data) {
              main_audiosend_next();
              goto main_repeat_audiosend;
            }
            audiosend_data = NULL;
            FD_CLR(audiosend_fd, &writepoll);
            fprintf(stderr, "Audio sent, waiting for sub-process to terminate\n");
            pclose(audiosend_proc);
            fprintf(stderr, "Audio sub-process terminated\n");
            if (ptt_flag && ioctl(ptt_fd, ptt_cmd_off, &ptt_flag) < 0) {
              return errno_error("Could not disable PTT");
            }
            if (!duplex && peertopeer) {
              fprintf(stderr, "Blocking TX for peer to respond\n");
              blocked_until = future(random_duration(mindelay, avrdelay));
              blocked = true;
            }
          }
        }
      }
      if (qam_demodulator_is_suspended(demodulator) && !audiosend_data) {
        qam_demodulator_run(demodulator, 0);  // NOTE: consumes pending data
        main_demodulator_reset();
      }

    }
  }
}
